//------------------------------------------------------------------------------
//	iso7816_vcc.c
//	Version 1.0, January 27, 2005
//	Copyright (c) 2005 by Matt Brunk
//	web:    http://www.smartcache.net/
//	e-mail: info@smartcache.net
//
//	The contents of this file are subject to the Artistic License (the 
//	"License").  You may not use this file except in compliance with the 
//	License.  You may obtain a copy of the License at: 
//	http://www.opensource.org/licenses/artistic-license.html
//------------------------------------------------------------------------------

//------------------------------------------------------------------------------
//	This source code implements an interface between an RS-232 serial port and 
//	an ISO 7816 smart card, and is targeted for the Atmel AT90S2313 
//	microcontroller.  The frequency of the clock signal applied to the Atmel 
//	microcontroller is assumed to be 9.216 MHz.  As a result, the 10 MHz 
//	version of the microcontroller must be used (e.g., AT90S2313-10PC).  Do 
//	not use the 4 MHz version of the microcontroller (e.g., AT90S2313-4PC)--if 
//	you do, you will be exceeding the absolute maximum rating by more than a 
//	factor of 2.  Atmel has this to say in their datasheet:
//
//	NOTICE:	Stresses beyond those listed under "Absolute Maximum Ratings"
//			may cause permanent damage to the device. This is a stress
//			rating only and functional operation of the device at these or
//			other conditions beyond those indicated in the operational 
//			sections of this specification is not implied. Exposure to 
//			absolute maximum rating conditions for extended periods may 
//			affect device reliability. 
//
//	Data is transmitted over the RS-232 serial port at a baud rate of 115,200 
//	bps.  The frequency of the clock signal applied to the ISO 7816 CLK contact
//	is assumed to be 4.608 MHz.  See the pin assignments below for other 
//	assumptions.
//
//  Timer/Counter1 (Timer1) is used in Pulse Width Modulation (PWM) mode. The
//	PWM output on pin PB3 (OC1) is input into an RC low-pass filter, then into
//	an operational amplifier configured as a voltage follower (unity gain). The
//	output of the operational amplifier is applied to the ISO VCC and CLK
//	contacts via a 4053 analog multiplexer/demultiplexer.
//
//	Timer/Counter0 is used for bit timings of data sent to and received from
//	the smart card. It is also used to signal the end of data transfers. The
//	prescale value for Timer0 determines which mode it is being used in.
//------------------------------------------------------------------------------

//------------------------------------------------------------------------------
//	Start of code.
//------------------------------------------------------------------------------
#include <inttypes.h>
#include <avr/eeprom.h>
#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/signal.h>

//------------------------------------------------------------------------------
//	Get rid of PC-lint warnings of no concern.
//------------------------------------------------------------------------------

//lint -esym(751,portb_enum)		not referenced
//lint -esym(751,portd_enum)		not referenced
//lint -esym(749,PB5_RTS_INV)		not referenced
//lint -esym(749,PD0_RXD)			not referenced
//lint -esym(749,PD7_NA)			not referenced
//lint -esym(754,pad1)				not referenced
//lint -esym(754,pad2)				not referenced
//lint -esym(754,pad3)				not referenced

//------------------------------------------------------------------------------
//	Constants
//------------------------------------------------------------------------------
static const uint8_t	BS1				= 0xA8;	// 1 bit spacing
static const uint8_t	BS1_5			= 0x80;	// 1.5 bit spacings
static const uint8_t	NUM_T0_COURSE	= 12;	// num T0 overflows at course
												// prescale for timeout
static const uint8_t	TCCR0_COARSE	= 5;	// used for timeouts
												// CK/1024 -- at 9.216 MHZ this
												// gives 28 ms
static const uint8_t	TCCR0_FINE		= 2;	// used for bit timings
												// CK/8 -- at 9.216 MHZ this
												// gives 222 us

//------------------------------------------------------------------------------
//	This type is used to specify the direction of data transfer in the Timer0 
//	interrupt vector.
//------------------------------------------------------------------------------
typedef enum
{
	STATE_RCV,			// from card
	STATE_XMIT			// to card
} state_enum;

//------------------------------------------------------------------------------
//	Device-specific pin assignments for PORT B.
//------------------------------------------------------------------------------
typedef enum
{
	PB0_WRITE_LED,		// anode (+) of write LED; cathode (-) of read LED
	PB1_READ_LED,		// anode (+) of read LED;  cathode (-) of write LED
	PB2_PIN14,			// used for bit timing debug
	PB3_PWM,			// output to op-amp for PWM
	PB4_ISO_SWITCH,		// ISO card switch; high means switch open and card in
	PB5_RTS_INV,		// RS-232 RTS inverted
	PB6_CTS,			// RS-232 CTS
	PB7_RTS				// RS-232 RTS used for RESET
} portb_enum;

//------------------------------------------------------------------------------
//	Device-specific pin assignments for PORT D.
//------------------------------------------------------------------------------
typedef enum
{
	PD0_RXD,			// UART receive data
	PD1_TXD,			// UART transmit data
	PD2_VCC_MUX,		// 4053 VCC MUX
	PD3_CLK_CTL,		// clock switch
	PD4_MUX_CTL,		// 4053 control
	PD5_ISO_RST,		// to ISO RST contact
	PD6_ISO_IO,			// to ISO I/O contact
	PD7_NA
} portd_enum;

//------------------------------------------------------------------------------
//	Variables with file scope.
//------------------------------------------------------------------------------
static uint8_t		G_buffer 		= 0;		// serial port data byte
static uint8_t		G_num_data_bits	= 0;		// number of bits transmitted
												// or received for current byte
												// of data
static uint8_t		G_parity_bit	= 0;		// parity bit for serial data
static state_enum	G_state			= STATE_RCV;// specifies ISO I/O direction
static uint8_t		G_t0_counter	= 0;		// count of Timer0 overflows
static uint8_t		G_xmit_char		= 0;		// byte to transmit to PC or card

//------------------------------------------------------------------------------
//	EEPROM data
//
//	The top value specifies the 8-bit value (the TOP value) that defines the 
//	pulse width used in PWM.  The pulse width (and consequently the duty 
//	cycle) is directly proportional to this value.  The larger the value, the 
//	wider the pulse width (and the larger the duty cycle).  The value of VCC 
//	applied to the ISO VCC and CLK contacts is directly proportional to this 
//	value.
//------------------------------------------------------------------------------

//------------------------------------------------------------------------------
typedef struct
{
	uint8_t	pad1;
	uint8_t	pad2;
	uint8_t	top;
	uint8_t	pad3;
} eeprom_struct;

static const eeprom_struct EE_PWM __attribute__((section(".eeprom"))) = 
{
	0,
	0,
	0xFF,	// top value; change this to change ISO VCC
	0
};

//------------------------------------------------------------------------------
//	Function prototypes
//------------------------------------------------------------------------------
static void debug_toggle_pin14(void);
static void handle_timer0_overflow_coarse(void);
static void init_ports(void);
static void init_pwm(void);
static void init_uart(void);
static void setup_for_start_bit(void);
static void turn_off_leds(void);
static void turn_on_read_led(void);
static void turn_on_write_led(void);
static void uart_transmit(void);

//------------------------------------------------------------------------------
//	The function main is effectively the reset handler.
//------------------------------------------------------------------------------
int main(void)
{
	cli();			// disable all interrupts while we init
	init_ports();	// initialize AVR ports
	init_pwm();		// initialize PWM
	init_uart();	// initialize the UART (RS-232 interface to PC)
	sei(); 			// reenable interrupts

	//--------------------------------------------------------------------------
	//	Loop forever.
	//--------------------------------------------------------------------------
	for ( ; ; )
	{		
		//----------------------------------------------------------------------
		//	Wait for card to be inserted.
		//----------------------------------------------------------------------
		loop_until_bit_is_set(PINB, PB4_ISO_SWITCH);
		
		//----------------------------------------------------------------------
		//	Card is inserted.  Wait for start of RESET signal (RS-232 RTS 
		//	line)  from serial port.
		//----------------------------------------------------------------------
		loop_until_bit_is_clear(PINB, PB7_RTS);
	
		//----------------------------------------------------------------------
		//	RESET is asserted.  Apply LOW to ISO RST contact to reset the card.
		//----------------------------------------------------------------------
		PORTD &= ~_BV(PD5_ISO_RST);

		//----------------------------------------------------------------------
		//	Wait for RS-232 RTS line to indicate end of RESET.
		//----------------------------------------------------------------------
		loop_until_bit_is_set(PINB, PB7_RTS);
		
		//----------------------------------------------------------------------
		//	Apply HIGH to ISO RST contact to reset the card.
		//----------------------------------------------------------------------
		PORTD |= _BV(PD5_ISO_RST);
		
		//----------------------------------------------------------------------
		//	Set up for the start bit (from first byte of ATR) from card.
		//----------------------------------------------------------------------
		setup_for_start_bit();		
	}
	
	return 0;	//lint !e527: Unreachable code at token 'return'
}

//------------------------------------------------------------------------------
//	This function initializes ports B and D.
//------------------------------------------------------------------------------
static void init_ports(void)
{
	//--------------------------------------------------------------------------
	//	Initialize port B.
	//--------------------------------------------------------------------------
	PORTB = 
		_BV(PB7_RTS)		|	// input: enable pull-up
		_BV(PB6_CTS)		|	// to RS-232 CTS
		_BV(PB4_ISO_SWITCH)	| 	// input: enable pull-up
		_BV(PB3_PWM)		| 	// always high
		_BV(PB1_READ_LED)	| 	// turn off LEDs
		_BV(PB0_WRITE_LED);
		

	DDRB =						// outputs
		_BV(PB6_CTS)		|
		_BV(PB3_PWM)		|
		_BV(PB2_PIN14)		|
		_BV(PB1_READ_LED)	|
		_BV(PB0_WRITE_LED);

	//--------------------------------------------------------------------------
	//	Initialize port D.
	//--------------------------------------------------------------------------
	PORTD = 
		_BV(PD6_ISO_IO)		|	// to ISO I/O
		_BV(PD5_ISO_RST)	|	// to ISO RST
		_BV(PD1_TXD);			// UART transmit data

	DDRD  =						// outputs 
		_BV(PD6_ISO_IO)		|
		_BV(PD5_ISO_RST)	|
		_BV(PD4_MUX_CTL)	|
		_BV(PD3_CLK_CTL)	|
		_BV(PD2_VCC_MUX)	|
		_BV(PD1_TXD);
}

//------------------------------------------------------------------------------
//	This function initializes the PWM output.
//------------------------------------------------------------------------------
static void init_pwm(void)
{
	uint16_t top;
	
	//---------------------------------------------------------------------------
	//	Load PWM value from EEPROM.
	//---------------------------------------------------------------------------
	top = eeprom_read_byte(&EE_PWM.top);

	//---------------------------------------------------------------------------
	//	Set the PWM value to the value just loaded from EEPROM.  Clear the 
	//	high byte, as we are only using 8-bit PWM.  Note that the PWM value is 
	//	set only here and never changes.  Consequently, the nominal voltage 
	//	applied to the ISO VCC contact never changes.
	//---------------------------------------------------------------------------
	OCR1A = top;

	//---------------------------------------------------------------------------
	//	Set up to use 8-bit, non-inverted PWM.  The PWM output is input into 
	//	an external RC low-pass filter and then into an OP-AMP configured as a 
	//	voltage follower.  The output of the OP-AMP is applied to the VCC and 
	//	CLK contacts of the smart card via the 4053 multiplexer.
	//---------------------------------------------------------------------------
	TCCR1A = _BV(COM1A1)|_BV(PWM10);
	TCCR1B = _BV(CS10);		// count at CK
}

//------------------------------------------------------------------------------
//	This function initializes the UART, the RS-232 interface to the PC.
//------------------------------------------------------------------------------
static void init_uart(void)
{
	//--------------------------------------------------------------------------
	//	Set baud rate; 115,200 at 9.216 MHz clock. See the datasheet.
	//--------------------------------------------------------------------------
	UBRR = 4;
	//--------------------------------------------------------------------------
	//	Enable receive, transmit, receive complete interrupt and 9-bit chars.
	//--------------------------------------------------------------------------
	UCR  = _BV(RXEN)|_BV(TXEN)|_BV(RXCIE)|_BV(CHR9);	
}

//------------------------------------------------------------------------------
//	This function turns off the read and write LEDs.
//------------------------------------------------------------------------------
static void turn_off_leds(void)
{
	//--------------------------------------------------------------------------
	//	Applying VCC to both anode (+) and cathode (-)  of LEDs turns them off.
	//--------------------------------------------------------------------------
	PORTB |= _BV(PB0_WRITE_LED);
	PORTB |= _BV(PB1_READ_LED);
}

//------------------------------------------------------------------------------
//	This function turns on the read LED.
//------------------------------------------------------------------------------
static void turn_on_read_led(void)
{
	PORTB &= ~_BV(PB0_WRITE_LED);
	PORTB |= _BV(PB1_READ_LED);
}

//------------------------------------------------------------------------------
//	This function turns on the write LED.
//------------------------------------------------------------------------------
static void turn_on_write_led(void)
{
	PORTB |= _BV(PB0_WRITE_LED);
	PORTB &= ~_BV(PB1_READ_LED);
}

//------------------------------------------------------------------------------
//	This function toggles the state of pin 14.  Pin 14 can be monitored with 
//	an oscilloscope to aid in development and debugging of this source code.
//------------------------------------------------------------------------------
static void debug_toggle_pin14(void)
{
	if ( bit_is_set(PINB, PB2_PIN14) )
	{
		PORTB &= ~_BV(PB2_PIN14);
	}
	else
	{
		PORTB |= _BV(PB2_PIN14);
	}
}

//------------------------------------------------------------------------------
//	This function sends a character over the UART (to the PC).  The 8-bit data 
//	to send must be in G_xmit_char and the parity bit in G_parity_bit.
//------------------------------------------------------------------------------
static void uart_transmit(void)
{
	if ( G_parity_bit )
	{
		UCR |= _BV(TXB8);
	}
	else
	{
		UCR &= ~_BV(TXB8);
	}
	
	UDR = G_xmit_char;	// transmit character over serial port
}

//------------------------------------------------------------------------------
//	This function sets up for the input capture interrupt to fire on the 
//	leading edge (falling) of the start bit from the card.
//------------------------------------------------------------------------------
static void setup_for_start_bit(void)
{
	//--------------------------------------------------------------------------
	//	Disable timer interrupts.
	//--------------------------------------------------------------------------
	TIMSK = 0;
	
	//--------------------------------------------------------------------------
	//	Set ISO I/O line to input.
	//--------------------------------------------------------------------------
	PORTD |= _BV(PD6_ISO_IO);
	DDRD  &= ~_BV(PD6_ISO_IO);
	PORTD &= ~_BV(PD6_ISO_IO);

	//--------------------------------------------------------------------------
	//	Set our state to READ.
	//--------------------------------------------------------------------------
	G_state = STATE_RCV;
	
	//--------------------------------------------------------------------------
	//	Clear timer/counter interrupt flags.
	//--------------------------------------------------------------------------
	TIFR = _BV(ICF1)|_BV(TOV0);
	
	//--------------------------------------------------------------------------
	//	Zero Timer0 counter.
	//--------------------------------------------------------------------------
	TCNT0 = 0;
	
	//--------------------------------------------------------------------------
	//	Set clock prescale for Timer0.
	//--------------------------------------------------------------------------
	TCCR0        = TCCR0_COARSE;
	G_t0_counter = NUM_T0_COURSE;	// # of Timer0 overflows at coarse prescale
									// before disabling Timer0.
	
	//--------------------------------------------------------------------------
	//	Set up the input capture to look for a falling slope, which will 
	//	represent the leading edge of the start bit.
	//--------------------------------------------------------------------------
	G_num_data_bits = 0;
	TIMSK			= _BV(TOIE0)|_BV(TICIE);
}

//------------------------------------------------------------------------------
//	This is the Timer0 overflow interrupt vector.  Timer0 is used to turn off 
//	the read LED and clear all timer interrupts.  Because Timer0 is only an 
//	8-bit value, its overflow period is too short for the duration needed to 
//	insure that all data transfers have completed.  Therefore, we use a 
//	counter (G_t0_counter) that gets decremented by 1 for each Timer0 overflow.
//	When the counter is 0, the entire wait period ends.
//------------------------------------------------------------------------------
static void handle_timer0_overflow_coarse(void)
{
	G_t0_counter--;
	if ( G_t0_counter == 0 )
	{
		turn_off_leds();				// turn off LEDs
		TCCR0 = 0;						// stop Timer0
		TIMSK = 0;						// disable all timer interrupts
		TIFR = _BV(ICF1)|_BV(TOV0);	// clear timer/counter interrupt flags
	}
	else
	{
		TCNT0 = 0;	// zero Timer0 counter
	}
}

//------------------------------------------------------------------------------
//	This is the UART receive character interrupt vector.  This is called when 
//	a character is received from the PC over the serial port.
//------------------------------------------------------------------------------
SIGNAL(SIG_UART_RECV)
{
	G_buffer = UDR;	// read serial data
	
	//--------------------------------------------------------------------------
	//	Save the value of the parity bit. We will be transmitting this to the
	//	card unchanged.
	//--------------------------------------------------------------------------
	G_parity_bit = bit_is_set(UCR, RXB8);
		
	//--------------------------------------------------------------------------
	//	Set Timer0 to count at rate used for bit timings.
	//--------------------------------------------------------------------------
	TCCR0 = TCCR0_FINE;
		
	//--------------------------------------------------------------------------
	//	Clear timer/counter interrupt flags.
	//--------------------------------------------------------------------------
	TIFR = _BV(ICF1)|_BV(TOV0);
	
	//--------------------------------------------------------------------------
	//	The only Timer interrupt we want now is Timer0 overflow (to transmit 
	//	data bits to the card).
	//--------------------------------------------------------------------------
	TIMSK = _BV(TOIE0);	// interrupt at TC0  (turn off T1 and input capture)
	
	//--------------------------------------------------------------------------
	//	Set the current value of Timer0 so that it overflows 1 bit spacing 
	//	from now (which is when the start bit ends and the first data bit 
	//	begins).
	//--------------------------------------------------------------------------
	TCNT0 = BS1;
	
	//--------------------------------------------------------------------------
	//	Set ISO I/O line direction to output.  We will be transmitting the 
	//	received byte to the card.
	//--------------------------------------------------------------------------
	PORTD |= _BV(PD6_ISO_IO);
	DDRD  |= _BV(PD6_ISO_IO);
	
	//--------------------------------------------------------------------------
	//	Clear ISO I/O line (this represents the leading edge of the start bit).
	//--------------------------------------------------------------------------
	PORTD &= ~_BV(PD6_ISO_IO);
	
	//--------------------------------------------------------------------------
	//	Set up our state.
	//--------------------------------------------------------------------------
	G_xmit_char 	= G_buffer;
	G_state			= STATE_XMIT;
	G_num_data_bits	= 0;
	
	//--------------------------------------------------------------------------
	//	Turn on the write LED.
	//--------------------------------------------------------------------------
	turn_on_write_led();	
}

//------------------------------------------------------------------------------
//	This is the Input Capture interrupt vector.  This is called at the leading 
//	edge of the start bit (data sent from card).
//------------------------------------------------------------------------------
SIGNAL(SIG_INPUT_CAPTURE1)
{
	G_buffer		= 0;
	G_num_data_bits = 0;
	TCCR0			= 0; // disable Timer0 counter

	//---------------------------------------------------------------------------
	//	Set up Timer0 prescale.
	//---------------------------------------------------------------------------
	TCCR0 = TCCR0_FINE;	

	//---------------------------------------------------------------------------
	//	Set up Timer0 to go off 1.5 bit spacings from now.  This gets us to 
	//	the middle of the first data bit (1 bit spacing for start bit plus 1/2 
	//	bit spacing for data bit).
	//---------------------------------------------------------------------------
	TIFR	= _BV(ICF1)|_BV(TOV0);	// clear timer interrupt flags
	TIMSK	= _BV(TOIE0);	// interrupt at TC0 (turn off T1 and input capture)
	TCNT0	= BS1_5;
	turn_on_read_led();
}

//------------------------------------------------------------------------------
//	This is the Timer0 overflow interrupt vector.
//------------------------------------------------------------------------------
SIGNAL(SIG_OVERFLOW0)
{
	//--------------------------------------------------------------------------
	//	The prescale value in TCCR0 is used to differentiate what Timer0 is 
	//	being used for.
	//--------------------------------------------------------------------------
	if ( TCCR0 != TCCR0_FINE )
	{
		handle_timer0_overflow_coarse();
	}
	else
	{	
		//----------------------------------------------------------------------
		//	Set up Timer0 to overflow 1 bit spacing later.
		//----------------------------------------------------------------------
		TCNT0 = BS1;
	
		//----------------------------------------------------------------------
		//	Toggle the state of pin 14.  This allows for bit transmit and 
		//	sample times to be monitored with an oscilloscope.
		//----------------------------------------------------------------------
		debug_toggle_pin14();
	
		G_num_data_bits++;
		
		switch ( G_state )
		{
			case STATE_XMIT:
			{
				//--------------------------------------------------------------
				//	Transmits to the card are done here.
				//--------------------------------------------------------------
				if ( G_num_data_bits <= 8 )
				{
					//----------------------------------------------------------
					//	Transmit the next data bit to the card.
					//----------------------------------------------------------
					if ( G_buffer & 0x01 )
					{
						PORTD |= _BV(PD6_ISO_IO);
					}
					else
					{
						PORTD &= ~_BV(PD6_ISO_IO);
					}
					
					G_buffer /= 2;	// shift buffer right by 1 bit		
				}
				else if ( G_num_data_bits == 9 )
				{
					//----------------------------------------------------------
					//	Transmit the parity bit to the card.
					//----------------------------------------------------------
					if ( G_parity_bit )
					{
						PORTD |= _BV(PD6_ISO_IO);
					}
					else
					{
						PORTD &= ~_BV(PD6_ISO_IO);
					}
				}
				else if ( G_num_data_bits == 10 )
				{
					//----------------------------------------------------------
					//	Transmit the stop bit to the card.
					//----------------------------------------------------------
					PORTD |= _BV(PD6_ISO_IO);
				}
				else if ( G_num_data_bits == 11 )
				{
					//----------------------------------------------------------
					//	Transmit of all bits to the card is complete.  Now 
					//	echo the character back to the serial port (PC) and 
					//	set up for possible data to be sent by the card.
					//----------------------------------------------------------
					uart_transmit();
					setup_for_start_bit();
				}
				else
				{
					//----------------------------------------------------------
					//	Should never get here.
					//----------------------------------------------------------
				}
			}
			break;
			
			case STATE_RCV:
			{
				//--------------------------------------------------------------
				//	Receives from the card are done here.
				//--------------------------------------------------------------
				uint8_t bit_val;
	
				bit_val = bit_is_set(PIND, PD6_ISO_IO);	// sample I/O line
				
				if ( G_num_data_bits <= 8 )
				{
					//----------------------------------------------------------
					//	This is a data bit.
					//----------------------------------------------------------
					if ( bit_val )
					{
						const uint8_t bit_num = (uint8_t)(G_num_data_bits - 1);//lint !e921
						G_buffer |= _BV(bit_num);//lint !e734 Loss of precision
					}
				}
				else if ( G_num_data_bits == 9 )
				{
					//----------------------------------------------------------
					//	This is the parity bit.  Transmit the data byte 
					//	received from the card to the serial port.
					//----------------------------------------------------------
					G_parity_bit = bit_val;				
					G_xmit_char  = G_buffer;
					uart_transmit();			
				}
				else if ( G_num_data_bits == 10 )
				{
					//----------------------------------------------------------
					//	This is the stop bit.  We're done, so setup for a 
					//	start bit from the card.
					//----------------------------------------------------------
					setup_for_start_bit();
				}
				else
				{
					//----------------------------------------------------------
					//	Should never get here.
					//----------------------------------------------------------
				}
			}
			break;
	
			default:
			{
				//--------------------------------------------------------------
				//	Should never get here.
				//--------------------------------------------------------------
			}
			break;
		}
	}
}

